function F = bvncdf(xvec,yvec,rho)
% cdf of the bivariate standard normal distribution with 
% correlation rho at the values in (xvec,yvec).
%
% Reference:
% O.A. Vasicek (1998), A Series Expansion for the Bivariate Normal Integral,
% KMV Corporation whitepaper.
F=zeros(length(xvec),length(yvec));
for xloop=1:length(xvec), for yloop=1:length(yvec)
  x=xvec(xloop); y=yvec(yloop);
  if x==0 && y==0
    c = -rho*sqrt(1/(1-rho^2)); F(xloop,yloop) = 1/4 - atan(c)/(2*pi);
    F(xloop,yloop) = 1/4 + asin(rho)/(2*pi);
  elseif y==0
    F(xloop,yloop)=mvncdf0(x,rho);
  elseif x==0
    F(xloop,yloop)=mvncdf0(y,rho);
  else
    F(xloop,yloop)=mvncdf0(x,(rho*x-y)/sqrt(x^2-2*rho*x*y+y^2)...
        *sign(x))+mvncdf0(y,(rho*y-x)/sqrt(x^2-2*rho*x*y+y^2)...
        *sign(y))-0.5*double(x*y<0);
  end
end, end

function F=mvncdf0(x,rho)
if rho^2 < 0.5 % Use tetrachoric series
  S=0;
  if rho~=0
    dS=1; k=0;
    while abs(dS)>1e-12
      k=k+1;
      dS=1/prod(1:k-1)/(2*k-1)*(-1)^(k-1)*2^(-k+1)...
          *Hermite(2*k-2,x)*rho^(2*k-1);
      S=S+dS;
    end
  end
  F=0.5*normcdf(x) + 1/(sqrt(2*pi))*normpdf(x)*S;
else % Use Vasicek (1998) series
  B=1/(2*pi)*sqrt(1-rho^2)*exp(-.5*(x^2/(1-rho^2)));
  A=-1/sqrt(2*pi)*abs(x)*normcdf(-abs(x)/sqrt(1-rho^2))+B;
  Q=A; k=0;
  while abs(A/Q)>1e-12
      k=k+1; B=(2*k-1)^2/(2*k*(2*k+1))*(1-rho^2)*B;
      A=-(2*k-1)/(2*k*(2*k+1))*x^2*A+B; Q=Q+A;
  end
  if rho>0
      F=min(normcdf(x),.5)-Q;
  else
      F=max(normcdf(x)-.5,0)+Q;
  end
end

function H=Hermite(k,x)
H=0;
for i=1:floor(k/2)+1
  H=H+prod(1:k)/prod(1:i-1)/prod(1:k-2*(i-1))...
      *(-1)^(i-1)*2^(-i+1)*x^(k-2*i+2);
end

function I=int(xvec,y,rho)
F=zeros(size(xvec));
mu=[0 0]'; si=[1 rho;rho 1];
for loop=1:length(xvec)
  x=xvec(loop); I(loop)=mvnpdf([x y]',mu,si);    
end
